import math
import json
import cv2
import os
from tqdm import tqdm
import numpy as np
import scipy.io as sio
import matplotlib.pyplot as plt
from scipy.spatial import cKDTree


def point_to_plane_distance(x, y, z, a, b, c):
    z_plane = a * x + b * y + c  
    distances = np.abs(z - z_plane) 
    return distances

def fit_plane(x, y, z):
    A = np.c_[x, y, np.ones(x.shape)]
    C, _, _, _ = np.linalg.lstsq(A, z, rcond=None)  
    return C  


def remove_noise(x, y, z, a, b, c, thresh=0.01):

    distances = point_to_plane_distance(x, y, z, a, b, c)
    mask = distances < thresh  
    mask1 = distances >= thresh
    x_keep = x[mask]
    y_keep = y[mask]
    z_keep = z[mask]
    z_mean = np.mean(z_keep)
    arr_keep = np.column_stack((x_keep,y_keep,z_keep))
    x_fliterd = x[mask1]
    y_fliterd = y[mask1]
    z_fliterd = z[mask1]
    arr_filterd = np.column_stack((x_fliterd,y_fliterd,z_fliterd))
   

    # fig = plt.figure(figsize=(10, 7))
    # ax = fig.add_subplot(111, projection='3d')
    # scatter = ax.scatter(x_filtered, y_filtered, z_filtered, c=z_filtered, cmap='viridis', marker='o')

    # cbar = plt.colorbar(scatter, ax=ax, shrink=0.5, aspect=5)
    # cbar.set_label('Z values')

    # ax.set_xlabel('X Axis')
    # ax.set_ylabel('Y Axis')
    # ax.set_zlabel('Z Axis')
    # ax.set_title('Filtered 3D Point Cloud (Outliers Removed)')

    # plt.show()
    return np.vstack([arr_keep, arr_filterd])

# def point_to_line_distance(point, A, B, C):
#     distance = abs(A * point[0] + B * point[1] + C) / np.sqrt(A**2 + B**2)
#     return distance

def get_support_points(x_filtered, y_filtered, z_filtered):
    support_points = []
    center_x = np.mean(x_filtered)
    center_y = np.mean(y_filtered)
    num_regions = 3
    angle_step = 2 * np.pi / num_regions  
    angles = np.arctan2(y_filtered - center_y, x_filtered - center_x) 
    distances = np.sqrt((x_filtered - center_x)**2 + (y_filtered - center_y)**2)  

    for i in range(num_regions):

        angle_min = -np.pi + i * angle_step
        angle_max = -np.pi + (i + 1) * angle_step
        region_mask = (angles >= angle_min) & (angles < angle_min + (2 * np.pi / 360 * 5))
        
        if np.any(region_mask):
            region_distances = distances[region_mask]
            region_idx = np.argmax(region_distances) #farthest points
            filtered_region_x = x_filtered[region_mask]
            filtered_region_y = y_filtered[region_mask]
            filtered_region_z = z_filtered[region_mask]
            point = [filtered_region_x[region_idx], filtered_region_y[region_idx], filtered_region_z[region_idx]]
            support_points.append(point)

    support_points = np.array(support_points)
    return support_points


def get_warpage(x, y, z):
    warpage = 0.0
    a, b, c = fit_plane(x, y, z)
    distances = (a * x + b * y - z + c) / np.sqrt(a**2 + b**2 + 1)
    max_distance = np.max(distances)
    min_distance = np.min(distances)
    warpage = max_distance - min_distance
    return warpage


# 定义函数：计算三维空间中点到直线的距离
def point_to_line_distance(point, line_point1, line_point2):
    line_vec = line_point2 - line_point1
    point_vec = point - line_point1
    distance = np.linalg.norm(np.cross(line_vec, point_vec)) / np.linalg.norm(line_vec)
    return distance


def get_mean_line_value(x, y, z, far_point_x, far_point_y, dist_thresh=0.5):
    
    a, b, c = fit_plane(x, y, z)
    x1,y1 = far_point_x, far_point_y
    x2,y2 = np.mean(x), np.mean(y)
    A = -(y2 - y1) if x2 != x1 else 1
    B = x2 - x1 if x2 != x1 else 0
    C = -(x2 * y1 - x1 * y2) if x2 != x1 else - y1
    dist_thresh = 0.5
    flat_data = np.column_stack((x, y))
    distances = np.abs(A * x + B * y + C) / np.sqrt(A**2 + B**2)
    intersections = flat_data[distances < dist_thresh]
    #print('len(intersections)= ', len(intersections))
    tree = cKDTree(flat_data)
    _, idx = tree.query(intersections, k=1)
    z_values = z[idx]

    z_plane_values = (a * intersections[:, 0] + b * intersections[:, 1] + c)
    warpage_diff = (z_values - z_plane_values)
    #print(warpage_diff)
    line_warpage = max(warpage_diff) - min(warpage_diff)
    
    distances_to_far_point = np.sqrt((intersections[:, 0] - far_point_x)**2 + (intersections[:, 1] - far_point_y)**2)

    # 找到距离最远的点索引
    max_distance_idx = np.argmax(distances_to_far_point)

    # 获取最远点的 x, y 坐标
    far_point_2 = intersections[max_distance_idx]

    # 构建 KD-Tree，便于快速查找最近邻
    tree = cKDTree(np.column_stack((x, y)))

    # 查找 far_point_x, far_point_y 最近的原始点
    _, idx1 = tree.query([far_point_x, far_point_y], k=1)
    far_point_1_z = z[idx1]  # 找到对应的 z 值

    # 查找 far_point_2 最近的原始点
    _, idx2 = tree.query(far_point_2, k=1)
    far_point_2_z = z[idx2]  # 找到对应的 z 值

    # 第一个端点的三维坐标
    far_point_1_3d = np.array([far_point_x, far_point_y, far_point_1_z])

    # 第二个端点的三维坐标
    far_point_2_3d = np.array([far_point_2[0], far_point_2[1], far_point_2_z])

    #print(far_point_1_3d, far_point_2_3d)

    # 查找中心点 (x2, y2) 最近的原始点，确定其 z 值
    _, idx_center = tree.query([round(x2), round(y2)], k=1)
    center_point_z = z[idx_center]  # 根据 (x2, y2) 最近点确定 z 值
    center_point_3d = np.array([round(x2), round(y2), center_point_z])
    
    line_bow = point_to_line_distance(center_point_3d, far_point_1_3d, far_point_2_3d)

    return line_warpage, line_bow, z_values


def get_theta(x, y, z, num_regions=60):
    
    center_x = np.mean(x)
    center_y = np.mean(y)
    angle_step = 2 * np.pi / num_regions  
    angles = np.arctan2(y - center_y, x - center_x) 
    distances = np.sqrt((x - center_x)**2 + (y - center_y)**2)  
    max_line_warpage = 0.0
    max_line_bow = 0.0
    max_warpage_theta = 0.0
    max_bow_theta = 0.0
    max_warpage_z = []
    max_bow_z = []
    for i in range(num_regions):
        angle_min = -np.pi + i * angle_step
        angle_max = -np.pi + (i + 1) * angle_step
        region_mask = (angles >= angle_min) & (angles < angle_max)
        if np.any(region_mask):
            region_distances = distances[region_mask]
            region_idx = np.argmax(region_distances) 
            far_point_x = x[region_mask][region_idx]
            far_point_y = y[region_mask][region_idx]
            mean_line_warpage, mean_line_bow, z_values = get_mean_line_value(x, y, z, far_point_x, far_point_y)
            if max_line_warpage < mean_line_warpage:
                max_line_warpage = mean_line_warpage
                max_warpage_theta = angle_min
                max_warpage_z = z_values
            if max_line_bow < mean_line_bow:
                max_line_bow = mean_line_bow
                max_bow_theta = angle_min
                max_bow_z = z_values
            #print('mean_line_warpage = ', mean_line_warpage)

    return max_line_warpage, max_line_bow, max_warpage_theta, max_bow_theta, max_warpage_z, max_bow_z


# 函数：将点映射到拟合椭圆上的对应点
def map_point_to_ellipse(contour_point, center, axes, angle):
    # 计算轮廓点相对于中心的偏移
    x_diff = contour_point[0][0] - center[0]
    y_diff = contour_point[0][1] - center[1]

    # 将轮廓点旋转至与椭圆一致的角度
    angle_rad = np.deg2rad(angle)
    x_rot = x_diff * np.cos(angle_rad) + y_diff * np.sin(angle_rad)
    y_rot = -x_diff * np.sin(angle_rad) + y_diff * np.cos(angle_rad)

    # 将点缩放到椭圆的长轴和短轴上
    x_ellipse = x_rot / axes[0]
    y_ellipse = y_rot / axes[1]

    # 映射点到椭圆上
    scaling_factor = 1 / np.sqrt(x_ellipse**2 + y_ellipse**2)
    x_ellipse_scaled = x_ellipse * scaling_factor * axes[0]
    y_ellipse_scaled = y_ellipse * scaling_factor * axes[1]

    # 旋转回原来的角度
    x_final = x_ellipse_scaled * np.cos(-angle_rad) + y_ellipse_scaled * np.sin(-angle_rad)
    y_final = -x_ellipse_scaled * np.sin(-angle_rad) + y_ellipse_scaled * np.cos(-angle_rad)

    # 返回映射后的椭圆点
    return (x_final + center[0], y_final + center[1])


def find_best_circle(image, min_thresh, max_thresh):
    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)   
    blurred = cv2.GaussianBlur(gray, (5, 5), 0)
    final_ellipse, final_thresh_img, final_contour = None, None, None
    final_error = 100
    final_bin_thresh = 100
    for bin_thresh in range(min_thresh, max_thresh, 5):
       
        _,  thresh_img = cv2.threshold(blurred, bin_thresh, 255, cv2.THRESH_BINARY)
        contours, _ = cv2.findContours(thresh_img, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        max_contour = max(contours, key=lambda contour: cv2.boundingRect(contour)[2] * cv2.boundingRect(contour)[3])
        ellipse = cv2.fitEllipse(max_contour)
        center = (int(ellipse[0][0]), int(ellipse[0][1]))  
        axes = (int(ellipse[1][0] / 2), int(ellipse[1][1] / 2))  
        angle = ellipse[2] 
        errors = []
        for point in max_contour:
            ellipse_point = map_point_to_ellipse(point, center, axes, angle)
            distance = np.linalg.norm(np.array(point) - np.array(ellipse_point))
       
            errors.append(distance)
        average_error = np.mean(errors)
     
        if average_error < final_error:
            final_error = average_error
            final_ellipse = ellipse 
            final_thresh_img = thresh_img
            final_bin_thresh = bin_thresh
            final_contour = max_contour
    return final_ellipse, final_thresh_img, final_bin_thresh, final_contour


def find_notch(white_path):

    image = cv2.imread(white_path)
    ellipse, thresh_img, bin_thresh, max_contour_ellipse = find_best_circle(image, 5, 50)
    
    background = np.zeros_like(thresh_img)
    cv2.ellipse(background, ellipse, (255, 255, 255), 1)  
    ellipse_points = np.column_stack(np.where(background == 255))
    (center_x, center_y), (axis_major, axis_minor), angle = ellipse
    
    height, width = thresh_img.shape
    min_gray = 10000
    min_pt = [0,0]  
    for pt in ellipse_points:
        x, y = pt[1], pt[0]
        #cv2.circle(image, (pt[1], pt[0]), 3, (0,255,255), 2)
        x_start = max(0, x - 5)
        x_end = min(width - 1, x + 5)
        y_start = max(0, y - 5)
        y_end = min(height - 1, y + 5)
        neighborhood = thresh_img[y_start:y_end+1, x_start:x_end+1]
        mean_value = np.mean(neighborhood)
        if mean_value < min_gray:
            min_gray = mean_value
            min_pt = (x, y)

    angle_rad = np.arctan2(min_pt[1] - center_y, min_pt[0] - center_x)

    thresh_img = cv2.drawContours(background, [max_contour_ellipse], -1, (255, 255, 255), thickness=cv2.FILLED)
   
    # cv2.circle(image, min_pt, 50, (0,255,255), 2)
    # #print(min_pt)
    # cv2.namedWindow('thresh_img', 0)
    # cv2.imshow('thresh_img', thresh_img)
    # cv2.namedWindow('contours', 0)
    # cv2.imshow('contours', image)
  
    # cv2.waitKey(0)
    return angle_rad, thresh_img

def get_eval_result(pointcloud_path, json_path, theta_notch):
    data = np.loadtxt(os.path.join(pointcloud_path), delimiter=',')
    
    x = data[:, 0]
    y = data[:, 1]
    z = data[:, 2]
    max_z = max(z)
    min_z = min(z)
    pv = max_z - min_z
    print('pv = ', pv)
    #a, b, c = fit_plane(x, y, z)
    #x_filtered, y_filtered, z_filtered = remove_noise(x, y, z, a, b, c, thresh=0.01)
    x_filtered, y_filtered, z_filtered = x, y, z
    
    support_points = get_support_points(x_filtered, y_filtered, z_filtered)
    a_bow, b_bow, c_bow = fit_plane(support_points[:,0], support_points[:,1], support_points[:,2])

    center_int = (round(np.mean(x_filtered)), round(np.mean(y_filtered)))
    #print('center_int = ', center_int)
    mask = (data[:, 0] == center_int[0]) & (data[:, 1] == center_int[1])
    z_center = data[mask, 2][0]

    fit_z_center = a_bow*center_int[0] + b_bow*center_int[1] + c_bow
    bow = (z_center - fit_z_center)
    print('global bow = ', bow)
    
    warpage = get_warpage(x_filtered, y_filtered, z_filtered)
    print('global warpage = ', warpage)

    max_line_warpage, max_line_bow, max_warpage_theta, max_bow_theta, max_warpage_z, max_bow_z = get_theta(x_filtered, y_filtered, z_filtered, 360)

    print('max_warpage_theta = ', max_warpage_theta)
    print('max_line_warpage = ', max_line_warpage)
    print('max_line_bow = ', max_line_bow)
    #print('max_warpage_z = ', max_warpage_z)

    x2 = int(round(np.mean(x)))  
    y2 = int(round(np.mean(y)))
    horizontal_mask = (y == y2)
    horizontal_z = z[horizontal_mask]  # 水平方向对应的 z 值
    horizontal_x = x[horizontal_mask]
   
    # 提取垂直方向的 z 值（即 x == x2 的所有点）
    vertical_mask = (x == x2)
    vertical_z = z[vertical_mask]  # 垂直方向对应的 z 值
    vertical_y = y[vertical_mask]
    
    top_x, top_y = x2, max(vertical_y)
    right_x, right_y = max(horizontal_x), y2
    #print(top_x, top_y)
    line_warpage_x, line_bow_x, z_values = get_mean_line_value(x, y, z, right_x, right_y)
    line_warpage_y, line_bow_y, z_values = get_mean_line_value(x, y, z, top_x, top_y)

    with open(json_path, 'w') as file:
        result_data = {
                'x':{
                    'height': horizontal_z.tolist(),
                    'bow':line_bow_x,
                    'warpage':line_warpage_x
                },
                'y':{
                    'height':vertical_z.tolist(),
                    'bow':line_bow_y,
                    'warpage':line_warpage_y
                },
                'max_bow':{
                    'angle':max_bow_theta,
                    'height':max_bow_z.tolist(),
                    'max_bow':max_line_bow
                },
                'max_warpage':{
                    'angle':max_warpage_theta,
                    'height':max_warpage_z.tolist(),
                    'max_warpage':max_line_warpage
                },
                'pv': pv,
                'notch_theta':theta_notch
        }
        #print(result_data)
        json.dump(result_data, file, indent=4)
    debug = False
    if debug:
        plt.scatter(x, y, c=z, cmap='viridis')
        plt.colorbar()
        plt.title('3D Scatter Plot of Reconstructed Surface')
        plt.xlabel('X Axis')
        plt.ylabel('Y Axis')

        fig = plt.figure(figsize=(10, 7))
        ax = fig.add_subplot(111, projection='3d')

        ax.scatter(x_filtered, y_filtered, z_filtered, color='green', label='Filtered Data', alpha=0.1)

        ax.scatter(support_points[:, 0], support_points[:, 1], support_points[:, 2], 
                color='red', marker='*', s=200, label='Support Points')
        ax.legend()

        ax.set_xlabel('X Axis')
        ax.set_ylabel('Y Axis')
        ax.set_zlabel('Z Axis')
        ax.set_title('3D Point Cloud with Uniformly Distributed Support Points')

        plt.show()