
import os
import matplotlib.pyplot as plt
from scipy.interpolate import griddata
from mpl_toolkits.mplot3d import Axes3D
import numpy as np
from scipy.sparse import lil_matrix
from scipy.sparse.linalg import spsolve
# 可视化结果（需要适当的x, y坐标映射）
import plotly.graph_objects as go
import plotly.io as pio
from scipy.spatial import Delaunay
from src.pcl_postproc import smooth_pcl, align2ref
from src.eval import fit_plane, remove_noise

def save_scatter_plot(x, y, z):
 
    # 创建三维散点图
    scatter = go.Scatter3d(
        x=x[:, 2],
        y=y[:, 2],
        z=z[:, 2],
        mode='markers',
        marker=dict(
            size=4,              # 点的大小
            color=z[:, 2],     # 使用 z 轴的值作为颜色映射
            colorscale='Viridis',  # 颜色映射
            opacity=0.8          # 点的透明度
        )
    )

    # 创建图表布局
    layout = go.Layout(
        title="Interactive 3D Scatter Plot",
        scene=dict(
            xaxis=dict(title="X-axis"),
            yaxis=dict(title="Y-axis"),
            zaxis=dict(title="Z-axis")
        )
    )

    # 生成图表
    fig = go.Figure(data=[scatter], layout=layout)

    # 保存为HTML文件，可以在浏览器中查看
    pio.write_html(fig, file='interactive_3d_scatter.html', auto_open=True)

def build_laplacian_matrix(points, triangles):
    N = len(points)
    A = lil_matrix((N, N))
    for tri in triangles:
        for i in range(3):
            A[tri[i], tri[i]] += 3  # 每个顶点累加对角线上的值
            for j in range(3):
                if i != j:
                    A[tri[i], tri[j]] -= 1  # 对应三角形顶点之间建立连接
    return A.tocsr()

def poisson_recons(x, y, gx, gy, debug=False):

    n = len(x)
    # theta = np.linspace(0, 2 * np.pi, n, endpoint=False)
    points = np.vstack((x, y)).T
    # Delaunay triangulation
    tri = Delaunay(points)
    triangles = tri.simplices

    # Build Laplacian matrix based on Delaunay triangulation
    L = build_laplacian_matrix(points, triangles)

    # Calculate divergence from synthetic gradients
    div_g = np.zeros(n)
    for i in range(n):
        div_g[i] = gx[i] + gy[i]  # Simple approximation

    # Solve the Poisson equation using the sparse matrix
    z_reconstructed = spsolve(L, div_g)

    # Color mapping based on the mean value of the z-coordinates at the vertices of each triangle
    # facecolors = z_reconstructed[triangles].mean(axis=1)
    return z_reconstructed

def reconstruction_cumsum(world_points, camera_points, screen_points, save_path, cam_id, debug=False, smooth=False, align=True, denoise=True):

    x = world_points[:, 0]
    y = world_points[:, 1]
    
    # 计算入射光线向量
    incident_vector = np.array([world_points[:, 0] - camera_points[:, 0],
                                world_points[:, 1] - camera_points[:, 1],
                                world_points[:, 2] - camera_points[:, 2]])

    incident_vector = incident_vector / np.linalg.norm(incident_vector, axis=0)

    # 计算反射光线向量
    reflection_vector = np.array([world_points[:, 0] - screen_points[:, 0],
                                world_points[:, 1] - screen_points[:, 1],
                                world_points[:, 2] - screen_points[:, 2]])
    
    reflection_vector = reflection_vector / np.linalg.norm(reflection_vector, axis=0)

    # 计算法向量
    normal_vector = incident_vector + reflection_vector
    normal_vector = normal_vector / np.linalg.norm(normal_vector, axis=0)
 
    normal_vector = np.nan_to_num(normal_vector, nan=0.0)
    # 计算法向量在X方向的梯度
    gradient_x = normal_vector[0]

    # 计算法向量在Y方向的梯度
    gradient_y = normal_vector[1]
    # gradient_x = np.zeros(gradient_x.shape)
    # gradient_y = np.zeros(gradient_x.shape)
    # import pdb
    # pdb.set_trace()
    # 计算z值，假设初始z值为0
    z = np.zeros(len(x))
    z[0] = 0  # 初始点的z值
    # import pdb; pdb.set_trace()
    if False:
    # 通过考虑梯度方向和位移计算每个点的z值
        for i in range(1, len(z)):
        # if i % 1000 == 0:
        # import pdb; pdb.set_trace()
            dx = x[i] - x[i-1]  # x方向的位移
            dy = y[i] - y[i-1]  # y方向的位移

            # 结合梯度和位移计算z的变化
            dz_x = gradient_x[i-1] * dx
            dz_y = gradient_y[i-1] * dy

            # dz_x = np.nan_to_num(dz_x, nan=0.0)
            # dz_y = np.nan_to_num(dz_y, nan=0.0)
            # print(dx, dy, dz_x, dz_y, z[i-1])
            # import pdb; pdb.set_trace()
            z[i] = z[i-1] + dz_x + dz_y
    # import pdb; pdb.set_trace()
        z = np.nan_to_num(z, nan=0.0)
    else:
        z = poisson_recons(x, y, gradient_x, gradient_y)
        # import pdb; pdb.set_trace()
    #print('-- point cloud: mean - {}; std - {}'.format(np.nanmean(z), np.nanstd(z)))
    if smooth:
        #print("-- point cloud is being smoothed.")
        smoothed_points = smooth_pcl(np.column_stack((x, y, z)))
        #print('-- nanmean after smoothing: mean - {}; std - {}'.format(np.nanmean(smoothed_points[:, 2]), np.nanstd(smoothed_points[:, 2])))
    if align:
        points_aligned = align2ref(smoothed_points)
        #print('-- nanmean after alignment: mean - {}; std - {}'.format(np.nanmean(points_aligned[:, 2]), np.nanstd(points_aligned[:, 2])))
    if denoise:
        z = points_aligned[:, 2]
        a, b, c = fit_plane(x, y, z)
        points_denoise = remove_noise(x, y, z, a, b, c, thresh=0.01)

    if debug:

        fig, axes = plt.subplots(1, 2, figsize=(12, 6))

        # 第一个子图
        cax0 = axes[0].scatter(x, y, c=gradient_x, cmap='viridis')
        axes[0].set_title('Visualization of x gradient')
        axes[0].set_xlabel('X Axis')
        axes[0].set_ylabel('Y Axis')
        fig.colorbar(cax0, ax=axes[0])

        cax1 = axes[1].scatter(x, y, c=gradient_y, cmap='viridis')
        axes[1].set_title('Visualization of y gradient')
        axes[1].set_xlabel('X Axis')
        axes[1].set_ylabel('Y Axis')
        fig.colorbar(cax1, ax=axes[1])
        plt.savefig(os.path.join(save_path, "gradient.png"))
        # 调整子图之间的间距
        plt.tight_layout()

        # 显示图像
        plt.show()


        # # 可视化点云
        plt.figure()
        # 绘制非规则点云
        plt.scatter(x, y, c=z, cmap='viridis')
        plt.colorbar()
        plt.title('3D Scatter Plot of Reconstructed Surface')
        plt.xlabel('X Axis')
        plt.ylabel('Y Axis')
        plt.savefig(os.path.join(save_path, str(cam_id) + "_recons.png"))
        # plt.zlabel('Y Axis')
        plt.show()

    # if smooth:
    # # # 可视化点云
    #     plt.figure()
    #     # 绘制非规则点云
    #     plt.scatter(x, y, c=smoothed_points[:, 2], cmap='viridis')
    #     plt.colorbar()
    #     plt.title('3D Scatter Plot of Smoothed points')
    #     plt.xlabel('X Axis')
    #     plt.ylabel('Y Axis')
    #     plt.savefig(os.path.join(save_path, "recons_smooth.png"))
    #     plt.close()
    #     # plt.zlabel('Y Axis')
    #     # plt.show()
    
    # if align:
    #     # # 可视化点云
    #     plt.figure()
    #     # 绘制非规则点云
    #     plt.scatter(x, y, c=points_aligned[:, 2], cmap='viridis')
    #     plt.colorbar()
    #     plt.title('3D Scatter Plot of Aligned points')
    #     plt.xlabel('X Axis')
    #     plt.ylabel('Y Axis')
    #     plt.savefig(os.path.join(save_path, "recons_aligned.png"))
    #     plt.close()
        # plt.zlabel('Y Axis')
        # plt.show()
    if False:

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        ax.scatter(x,y,z, cmap='viridis')
        ax.set_title('Visualization of screen points')
        ax.set_xlabel('X axis')
        ax.set_ylabel('Y axis')
        ax.set_zlabel('Z axis')
        plt.show()

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        ax.scatter(x,y,smoothed_points[:, 2], cmap='viridis')
        ax.set_title('3D Scatter Plot of Smoothed points')
        ax.set_xlabel('X axis')
        ax.set_ylabel('Y axis')
        ax.set_zlabel('Z axis')
        plt.show()

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        ax.scatter(x, y, c=points_aligned[:, 2], cmap='viridis')
        ax.set_title('3D Scatter Plot of Aligned points')
        ax.set_xlabel('X axis')
        ax.set_ylabel('Y axis')
        ax.set_zlabel('Z axis')
        plt.close()

    # return points_aligned[:,2]
    return z, smoothed_points[:,2], points_aligned[:,2], points_denoise
