import cv2
import os
from tqdm import tqdm
import numpy as np
import scipy.io as sio
import matplotlib.pyplot as plt
from scipy.interpolate import griddata
import glob


def read_and_undis(img_path, mtx, dist):

    img = cv2.imread(img_path)
    dst = cv2.undistort(img, mtx, dist, None, mtx)
    gray = cv2.cvtColor(dst, cv2.COLOR_BGR2GRAY)
    return gray

def write_point_cloud(file_path, x, y, z):
    # 将 x, y, z 列组合成一个二维数组
    point_cloud = np.column_stack((x, y, z))
    # import pdb; pdb.set_trace()
    # 将数组写入文本文件，使用空格作为分隔符
    np.savetxt(file_path, point_cloud, fmt='%.10f', delimiter=',')
    print(f"Point cloud saved to {file_path}")

def get_baseline_params(filename):
    world_raw = sio.loadmat(filename)
    return world_raw

def binarize(image, threshold=85):
    # Read and binarize the image
    # image = cv2.imread(img_path, cv2.IMREAD_GRAYSCALE)
    _, binary_image = cv2.threshold(image, threshold, 255, cv2.THRESH_BINARY)
    num_labels, labels, stats, centroids = cv2.connectedComponentsWithStats(binary_image, connectivity=8)

    # We assume that the largest connected component is the object of interest (the large circle)
    # Identify the label with the largest area, ignoring the background label (label 0)
    largest_label = 1 + np.argmax(stats[1:, cv2.CC_STAT_AREA])

    # Create a mask for the largest connected component
    filtered_image = np.zeros_like(binary_image)
    filtered_image[labels == largest_label] = 255

    return filtered_image

def get_ellipse_points(binary, contours, grid_spacing, ellipse_center, ellipse_axes, angle_rad):
    image_color = cv2.cvtColor(binary, cv2.COLOR_GRAY2BGR)
    grid_points = []
    
    h, w = binary.shape[:2]
    y_indices, x_indices = np.mgrid[0:h:grid_spacing, 0:w:grid_spacing]

    # 将坐标转换为椭圆坐标系
    x_shifted = x_indices - ellipse_center[0]
    y_shifted = y_indices - ellipse_center[1]

    # 旋转坐标系
    cos_angle = np.cos(angle_rad)
    sin_angle = np.sin(angle_rad)
    x_rotated = x_shifted * cos_angle + y_shifted * sin_angle
    y_rotated = -x_shifted * sin_angle + y_shifted * cos_angle

    # 判断每个点是否在椭圆内部或边界上
    ellipse_eq = (x_rotated**2 / ellipse_axes[0]**2) + (y_rotated**2 / ellipse_axes[1]**2)
    inside_ellipse = ellipse_eq < 1

    # 记录并绘制在椭圆内部或边界上的点
    grid_points = []
    contour_points = []

    count = 0
    for i in range(len(contours)):
        for point in contours[i]:
            # import pdb; pdb.set_trace()
            count += 1
            assert point.shape == (1, 2), point.shape
            if count == 400:
                cv2.circle(image_color, point[0], 1, (0, 255, 0), -1)
                cv2.putText(image_color, str(point[0]), (point[0][0], point[0][1]), cv2.FONT_HERSHEY_SIMPLEX, 0.75, (0, 255, 0), 2)
            else:
                cv2.circle(image_color, point[0], 1, (0, 0, 255), -1)

            contour_points.append(point[0].tolist())
    
    return grid_points, contour_points, image_color

def fit_circle(contour_points, grid_spacing):
    z_contour_points = contour_points[:, 2]
    points_float = np.array(contour_points[:, :2], dtype=np.float32)
    # import pdb; pdb.set_trace()
    (center, radius) = cv2.minEnclosingCircle(points_float)

    print(f"Fitted Circle Center: {center}")
    print(f"Fitted Circle Radius: {radius}")

    # Create a grid of points that covers the bounding box of the circle
    x_min = int(center[0] - radius)
    x_max = int(center[0] + radius)
    y_min = int(center[1] - radius)
    y_max = int(center[1] + radius)
    # Generate grid points within the bounding box
    x_vals = np.arange(x_min, x_max, grid_spacing)
    y_vals = np.arange(y_min, y_max, grid_spacing)
    x_grid, y_grid = np.meshgrid(x_vals, y_vals)
    grid_points = np.vstack([x_grid.ravel(), y_grid.ravel()]).T

    # Calculate the squared distances of the grid points from the center
    distances_squared = (grid_points[:, 0] - center[0])**2 + (grid_points[:, 1] - center[1])**2

    # Filter the points to keep only those within the circle
    inside_circle = distances_squared <= radius**2
    grid_points_inside_circle = grid_points[inside_circle]
    z_col = np.full((len(grid_points_inside_circle), 1), z_contour_points[0])

    return np.column_stack((grid_points_inside_circle, z_col))


def fit_sector(contour_points, grid_spacing):
    z_contour_points = contour_points[:, 2]
    # 提取 xy 平面上的轮廓点
    contour_xy = contour_points[:, :2]
    # 确保 contour_xy 是 numpy 数组
    contour_xy = np.array(contour_xy, dtype=np.int32)
    # 获取轮廓的边界框
    x, y, w, h = cv2.boundingRect(contour_xy)

    # 生成 xy 平面上的网格点
    xx, yy = np.meshgrid(np.arange(x, x + w, grid_spacing), np.arange(y, y + h, grid_spacing))
    grid_points_xy = np.vstack([xx.ravel(), yy.ravel()]).T

    # 过滤出轮廓内部的点，并记录它们的索引
    inside_points_xy = []
    inside_indices = []
    for idx, point in enumerate(grid_points_xy):
        # 使用cv2.pointPolygonTest判断点是否在轮廓内部
        if cv2.pointPolygonTest(contour_xy, (float(point[0]), float(point[1])), False) >= 0:
            inside_points_xy.append(point)
            inside_indices.append(idx)

    inside_points_xy = np.array(inside_points_xy)

    # 使用过滤后的索引获得网格点，并为每个点设置 z 值
    grid_points_inside_sector = grid_points_xy[inside_indices]
    z_col = np.full((len(grid_points_inside_sector), 1), z_contour_points[0])

    return np.column_stack((grid_points_inside_sector, z_col))
    

    #return inside_points_xyz

# def get_meshgrid_contour(binary, grid_spacing, save_path=None, debug=False):
#     # Detect contours
#     contours, _ = cv2.findContours(binary, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)

#     # 找到最大轮廓
#     max_index, max_contour = max(enumerate(contours), key=lambda x: cv2.boundingRect(x[1])[2] * cv2.boundingRect(x[1])[3])

#     # 获取轮廓的边界框
#     x, y, w, h = cv2.boundingRect(max_contour)

#     # 生成网格点
#     xx, yy = np.meshgrid(np.arange(x, x + w, grid_spacing), np.arange(y, y + h, grid_spacing))
#     grid_points = np.vstack([xx.ravel(), yy.ravel()]).T

#     # 过滤出轮廓内部的点
#     inside_points = []
#     for point in grid_points:
#         # 使用cv2.pointPolygonTest判断点是否在轮廓内部
#         if cv2.pointPolygonTest(max_contour, (float(point[0]), float(point[1])), False) >= 0:
#             inside_points.append(point)

#     inside_points = np.array(inside_points)

#     # 可视化
#     if debug:
#         plt.figure(figsize=(8, 8))
#         plt.imshow(binary, cmap='gray')
#         plt.scatter(inside_points[:, 0], inside_points[:, 1], color='blue', s=1, label='Grid Points Inside Contour')
#         plt.plot(max_contour[:, 0, 0], max_contour[:, 0, 1], color='red', linewidth=2, label='Contour')
#         plt.legend()
#         plt.title('Meshgrid Points Inside Contour')
#         if save_path:
#             plt.savefig(save_path)
#         plt.show()

#     return inside_points

def get_meshgrid_contour(binary, grid_spacing, save_path, debug=False):

    # img_path = 'imgs/temp/white.bmp'
    # image = cv2.imread(img_path, cv2.IMREAD_GRAYSCALE)
    # binary = binarize(image)
    # Detect contours
    kernel = np.ones((5, 5), np.uint8)

    # 应用腐蚀操作
    erosion = cv2.erode(binary, kernel, iterations=1)
    contours1, _ = cv2.findContours(binary, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
    contours, _ = cv2.findContours(erosion, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)

    max_index, max_contour = max(enumerate(contours), key=lambda x: cv2.boundingRect(x[1])[2] * cv2.boundingRect(x[1])[3])
    # max_index, max_contour1 = max(enumerate(contours1), key=lambda x: cv2.boundingRect(x[1])[2] * cv2.boundingRect(x[1])[3])

    # output_image = cv2.cvtColor(erosion, cv2.COLOR_GRAY2BGR)  # 转换为彩色图像用于绘制
    # cv2.drawContours(output_image, [max_contour1], -1, (255, 255, 0), thickness=2)
    # cv2.namedWindow('output_image', 0)
    # cv2.imshow('output_image', output_image)
    # cv2.waitKey(0)
    #print('max_contour = ',max_contour[0])
    max_contour_lst = []
    for pt in max_contour:
        #print('[pt[0][1],pt[0][0]] = ', [int(pt[0][1]), int(pt[0][0])])
        #print('pt[0].tolist() =', pt[0].tolist())
        #max_contour_lst.append([int(pt[0][1]), int(pt[0][0])])
        max_contour_lst.append(pt[0].tolist())
    #print('max_contour_lst = ',max_contour_lst)
    max_contour_array = np.array(max_contour_lst)

    return max_contour_array

    # Initialize ellipse
    # ellipse = None

    # # Check and fit an ellipse
    # if contours:
    #     sorted_contours = sorted(contours, key=cv2.contourArea, reverse=True)
    #     # import pdb; pdb.set_trace()
    #     for contour in sorted_contours:
    #         if len(contour) >= 5:  # Need at least 5 points to fit an ellipse
    #             ellipse = cv2.fitEllipse(contour)
    #             break
    #     else:
    #         print("没有足够的轮廓点来拟合椭圆")
    # else:
    #     print("没有检测到足够的轮廓或轮廓点太少")

    # # Continue with your grid generation and other tasks if ellipse was successfully fitted
    # if ellipse:
    #     # Convert binary image to BGR for color drawing
    #     grid_points = []
    #     ellipse_center = ellipse[0]
    #     ellipse_axes = (ellipse[1][0] / 2, ellipse[1][1] / 2)  # Dividing axes by 2 to get semi-axes
    #     angle_rad = np.deg2rad(ellipse[2])
    #     # Coordinate of the grid points
    #     grid_points, contour_points, image_color = get_ellipse_points(binary, contours, grid_spacing, ellipse_center, ellipse_axes, angle_rad)   
    # else:
    #     raise ValueError("未能拟合椭圆")

    # if debug:
    #     # print('There are {} points in the ellipse'.format(len(grid_points)))
    #     # print('Mean: {}'.format(np.mean(np.array(grid_points))))
    #     print('There are {} points on the ellipse'.format(len(contour_points)))
    #     print('Mean: {}'.format(np.mean(np.array(contour_points))))
    #     cv2.ellipse(image_color, ellipse, (255,0,255), 2)
    #     cv2.imwrite(os.path.join(save_path, "meshgrid.png"), image_color)
    #     cv2.namedWindow('meshgrid', 0)
    #     cv2.imshow('meshgrid', image_color)
    #     cv2.waitKey(0)
    #     cv2.destroyAllWindows()
        
    # # grid_points = np.array(grid_points)
    # contour_points = np.array(contour_points)
    # # print('x:', grid_points[:, 0].max()-grid_points[:, 0].min())
    # # print('y:', grid_points[:, 1].max()-grid_points[:, 1].min())
    # return grid_points, contour_points
    # # return np.fliplr(grid_points), np.fliplr(contour_points)


def get_meshgrid(binary, grid_spacing, save_path, debug=False):

    # img_path = 'imgs/temp/white.bmp'
    # image = cv2.imread(img_path, cv2.IMREAD_GRAYSCALE)
    # binary = binarize(image)

    # Detect contours
    contours, _ = cv2.findContours(binary, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
    # 初始化最大范围的变量
    x_min, y_min = float('inf'), float('inf')
    x_max, y_max = 0, 0

    # 计算所有轮廓的最大范围
    for contour in contours:
        x, y, w, h = cv2.boundingRect(contour)
        x_min = min(x_min, x)
        y_min = min(y_min, y)
        x_max = max(x_max, x + w)
        y_max = max(y_max, y + h)
    # 存储采样点的列表
    sampled_points = []

    # 在最大边界框范围内进行采样，步长为 grid_spacing
    for i in range(x_min, x_max, grid_spacing):
        for j in range(y_min, y_max, grid_spacing):
            # 遍历所有轮廓，检查该点是否在任意轮廓内部
            for contour in contours:
                if cv2.pointPolygonTest(contour, (i, j), False) >= 0:
                    sampled_points.append((i, j))
                    break  # 一旦找到在某个轮廓内，就不需要再检查其他轮廓了
    #print('sampled_points ',sampled_points)
    grid_points = np.array((sampled_points))
    grid_points = grid_points[:, ::-1]
    # 将采样点显示在图像上
    output_image = cv2.cvtColor(binary, cv2.COLOR_GRAY2BGR)  # 转换为彩色图像用于绘制
    for point in sampled_points:
        cv2.circle(output_image, point, 2, (0, 0, 255), -1)  # 画小圆点

    # 显示结果
    cv2.namedWindow('Sampled Points', 0)
    cv2.imshow("Sampled Points", output_image)
    cv2.waitKey(0)
    cv2.destroyAllWindows()
    return grid_points

    # # Initialize ellipse
    # ellipse = None

    # # Check and fit an ellipse
    # if contours:
    #     sorted_contours = sorted(contours, key=cv2.contourArea, reverse=True)
    #     # import pdb; pdb.set_trace()
    #     for contour in sorted_contours:
    #         if len(contour) >= 5:  # Need at least 5 points to fit an ellipse
    #             ellipse = cv2.fitEllipse(contour)
    #             break
    #     else:
    #         print("没有足够的轮廓点来拟合椭圆")
    # else:
    #     print("没有检测到足够的轮廓或轮廓点太少")
    # print('ellipse = ', ellipse)
    # # Continue with your grid generation and other tasks if ellipse was successfully fitted
    # if ellipse:
    #     # Convert binary image to BGR for color drawing
    #     grid_points = []
    #     ellipse_center = ellipse[0]
    #     ellipse_axes = (ellipse[1][0] / 2, ellipse[1][1] / 2)  # Dividing axes by 2 to get semi-axes
    #     angle_rad = np.deg2rad(ellipse[2])
    #     # Coordinate of the grid points
    #     grid_points, contour_points, image_color = get_ellipse_points(binary, contours, grid_spacing, ellipse_center, ellipse_axes, angle_rad)   
    # else:
    #     raise ValueError("未能拟合椭圆")
    # cv2.namedWindow('meshgrid', )
    # cv2.imshow('meshgrid', binary)
    # cv2.waitKey(0)
    # cv2.destroyAllWindows()

    # if debug:
    #     # print('There are {} points in the ellipse'.format(len(grid_points)))
    #     # print('Mean: {}'.format(np.mean(np.array(grid_points))))
    #     print('There are {} points on the ellipse'.format(len(contour_points)))
    #     print('Mean: {}'.format(np.mean(np.array(contour_points))))
    #     cv2.imwrite(os.path.join(save_path, "meshgrid.png"), image_color)
    #     cv2.imshow('meshgrid', image_color)
    #     cv2.waitKey(0)
    #     cv2.destroyAllWindows()
        
    # # grid_points = np.array(grid_points)
    # contour_points = np.array(contour_points)
    # # print('x:', grid_points[:, 0].max()-grid_points[:, 0].min())
    # # print('y:', grid_points[:, 1].max()-grid_points[:, 1].min())
    # return grid_points, contour_points
    # # return np.fliplr(grid_points), np.fliplr(contour_points)

def affine_img2world(grid_points, A, Rc2w, Tc2w, d):
     # 计算交点
    intersection_points_world = []

    for point in grid_points:
        x_img, y_img = point[0], point[1]
        ray_camera_coordinate = np.array([
            (x_img - A[0, 2]) / A[0, 0],
            (y_img - A[1, 2]) / A[1, 1],
            1
        ])
        ray_camera_coordinate /= np.linalg.norm(ray_camera_coordinate)  # 归一化
        ray_world_coordinate = Rc2w @ ray_camera_coordinate
        cameraCenter_world_coordinate = Tc2w

        # 计算世界坐标系中的交点
        t = (-d - cameraCenter_world_coordinate[2]) / ray_world_coordinate[2]
        intersection_point = cameraCenter_world_coordinate.T + t * ray_world_coordinate
        intersection_points_world.append(intersection_point)

    # 输出世界坐标系中的交点
    intersection_points_world = np.array(intersection_points_world)
    intersection_points_world = intersection_points_world.reshape(-1, 3)

    # x_w_data = np.column_stack((grid_points, intersection_points_world[:, 0]))
    # y_w_data = np.column_stack((grid_points, intersection_points_world[:, 1]))
    # z_w_data = np.column_stack((grid_points, intersection_points_world[:, 2]))

    # return [x_w_data, y_w_data, z_w_data]
    return intersection_points_world

def get_world_points(contour_points, world_mat_contents, grid_spacing, d, save_path, debug=False):
    
    A = world_mat_contents['camera_matrix']
    Rw2c = world_mat_contents['rotation_matrix']
    Rc2w = np.linalg.inv(Rw2c)
    Tw2c = world_mat_contents['translation_vector']
    Tc2w = -np.dot(np.linalg.inv(Rw2c), Tw2c)
    # import pdb; pdb.set_trace()
    # world_points = affine_img2world(grid_points, A, Rc2w, Tc2w, d)
    
    #print('contour_points = ', np.max(contour_points), np.min(contour_points))
    world_points_contours = affine_img2world(contour_points, A, Rc2w, Tc2w, d)
    #world_points_fit = fit_circle(world_contour_points, grid_spacing)
    world_points_fit = fit_sector(world_points_contours, grid_spacing)
    # assert np.abs(world_contour_points[:, 0].max()-world_contour_points[:, 0].min()-(world_contour_points[:, 1].max()-world_contour_points[:, 1].min())) < 1
    # import pdb; pdb.set_trace()
    if debug:
        # print('world x:', world_points_fit[:, 0].max()-world_points_fit[:, 0].min())
        # print('world y:', world_points_fit[:, 1].max()-world_points_fit[:, 1].min())

        #print('world contour x:', world_contour_points[:, 0].max()-world_contour_points[:, 0].min())
        #print('world contour y:', world_contour_points[:, 1].max()-world_contour_points[:, 1].min())

        plt.figure()
        
        plt.scatter(world_points_fit[:, 0], world_points_fit[:, 1],  color='red', alpha=0.5)
        #plt.scatter(world_points_fit[:, 0], world_points_fit[:, 1], color='green', alpha=0.5)
        plt.title('world points fit')
        plt.xlabel('X Axis')
        plt.ylabel('Y Axis')

        # 调整子图之间的间距
        plt.tight_layout()
        plt.savefig(os.path.join(save_path, "world_points.png"))
        # 显示图像
        plt.show()
    # import pdb; pdb.set_trace()
    # return np.concatenate((world_points_fit, world_contour_points), axis=0)
    return world_points_fit



def get_camera_points(world_points, world_mat_contents, save_path, cam_id, debug=False):
    
    x_w_data, y_w_data, z_w_data = world_points[:, 0], world_points[:, 1], world_points[:, 2]
    A = world_mat_contents['camera_matrix']
    Tw2c = world_mat_contents['translation_vector']
    Rw2c = world_mat_contents['rotation_matrix']
    Rc2w = np.linalg.inv(Rw2c)
    Tc2w = -np.dot(Rc2w, Tw2c)

    # 初始化数据数组
    x_c_array, y_c_array, z_c_array, u_p_array, v_p_array = [], [], [], [], []

    # 确保x_w, y_w, z_w具有相同形状
    assert x_w_data.shape == y_w_data.shape == z_w_data.shape
    #print(x_w_data.shape)
    for i in range(x_w_data.shape[0]):
        x_w, y_w, z_w = x_w_data[i], y_w_data[i], z_w_data[i]

        world_point_now = np.array([x_w, y_w, z_w])
        picture_point_now = A @ (Rw2c @ world_point_now[:, np.newaxis] + Tw2c)
        picture_point_now /= picture_point_now[2]

        u_p = np.round(picture_point_now[0]).item()  # 使用.item()将单元素数组转换为标量
        v_p = np.round(picture_point_now[1]).item()

        # 计算相机坐标
        x_c, y_c, z_c = Tc2w.ravel()
        x_c_array.append(x_c)
        y_c_array.append(y_c)
        z_c_array.append(z_c)
        u_p_array.append(u_p)
        v_p_array.append(v_p)
    
    x_c_array, y_c_array, z_c_array, u_p_array, v_p_array = np.array(x_c_array), np.array(y_c_array), np.array(z_c_array), np.array(u_p_array), np.array(v_p_array)
    camera_points = np.column_stack((x_c_array, y_c_array, z_c_array))

    if debug:
        print('u_p max min:', u_p_array.max(),u_p_array.min())
        print('v_p max min:', v_p_array.max(), v_p_array.min())
        print('u_p:', u_p_array.max()-u_p_array.min())
        print('v_p:', v_p_array.max()-v_p_array.min())

        plt.figure()
        plt.scatter(u_p_array, v_p_array, alpha=0.5)
        plt.title('camera points')
        plt.xlabel('X Axis')
        plt.ylabel('Y Axis')

        # 调整子图之间的间距
        plt.tight_layout()
        plt.savefig(os.path.join(save_path, str(cam_id) + "_camera_points.png"))

        # 显示图像
        plt.show()

    return camera_points, u_p_array, v_p_array

def get_screen_points(point_data, x_phase_unwrapped, y_phase_unwrapped, screen_params, screen_to_world, cfg, save_path, cam_id, debug=False):
    
    # x_phase_unwrapped = x_phase_unwrapped.T
    # y_phase_unwrapped = y_phase_unwrapped.T
    # import pdb; pdb.set_trace()

    screenPixelSize = np.array(cfg['screenPixelSize'])
    Axis_conversion = np.array(cfg['Axis_conversion'][cam_id])
    screenorigin_point = np.array(cfg['screenorigin_point'])

    # 提取旋转矩阵和平移向量
    Rs2w = screen_to_world['screen_to_world_rotation']
    Ts2w = screen_to_world['screen_to_world_translation']

    # 计算 x_phase_unwrapped 中所有非 NaN 值的平均值
    average_x_phase_unwrapped = np.nanmean(x_phase_unwrapped)
    average_y_phase_unwrapped = np.nanmean(y_phase_unwrapped)

    # import pdb; pdb.set_trace()
    # 计算 value_map 并找到最小值
    value_map = np.abs(x_phase_unwrapped - average_x_phase_unwrapped) + np.abs(y_phase_unwrapped - average_y_phase_unwrapped)
    min_index = np.unravel_index(np.nanargmin(value_map), value_map.shape)
    
    # import pdb; pdb.set_trace()
    # 选定一个中心相位值作为绝对相位值
    # abs_phasepoint_picture = np.array(min_index)
    abs_phasepoint_x = x_phase_unwrapped[min_index]
    abs_phasepoint_y = y_phase_unwrapped[min_index]
    abs_phasepoint = np.array([abs_phasepoint_x, abs_phasepoint_y])
    # print('----------------------------------------')
    # print("abs_phasepoint", np.mean(np.isnan(abs_phasepoint)))
    # 根据相位值求解在屏幕上的像素点位置
    abs_phasepoint_screen = (abs_phasepoint / (2 * np.pi)) * (screenPixelSize/ cfg['k'])
    # print("abs_phasepoint_screen", np.mean(np.isnan(abs_phasepoint_screen)))

    # 利用相对位置信息确定绝对相位点的世界坐标系位置
    arrow_abs_mm = np.array([Axis_conversion[0] * cfg['pixelSize_mm'] * (abs_phasepoint_screen[0] - screenorigin_point[0]),
                            Axis_conversion[1] * cfg['pixelSize_mm'] * (abs_phasepoint_screen[1] - screenorigin_point[1]),
                            0])
    abs_phasepoint_world = np.dot(Rs2w, arrow_abs_mm) + Ts2w.T
    
    # print('abs_phasepoint 绝对相位值:')
    # print(abs_phasepoint)
    # print('abs_phasepoint_world 绝对相位值对应的世界坐标系的位置:')
    # print(abs_phasepoint_world)

    # print('----------------------------------------')

    x_data, y_data, z_data = [], [], []

    for idx in range(len(point_data['u_p'])):
        # i, j, u_now = point_data['u_p'][idx]
        u_now = point_data['u_p'][idx]
        v_now = point_data['v_p'][idx]
        u_now, v_now = int(u_now), int(v_now)
       
        phase_x_now = x_phase_unwrapped[v_now, u_now]  # 注意这里的索引要按照Python的行列顺序，与MATLAB相反 raw
        phase_y_now = y_phase_unwrapped[v_now, u_now]

        #phase_x_now = x_phase_unwrapped[u_now, v_now]  
        #phase_y_now = y_phase_unwrapped[u_now, v_now]

        # import pdb; pdb.set_trace()
        derta_phase_x = phase_x_now - abs_phasepoint[0]
        derta_phase_y = phase_y_now - abs_phasepoint[1]
        #assert not np.isnan(derta_phase_x).any(), "derta_phase_x"
        #assert not np.isnan(derta_phase_y).any(), "derta_phase_y"
        derta_mm = (np.array([derta_phase_x, derta_phase_y]) / (2 * np.pi)) * (screenPixelSize / cfg['k']) * cfg['pixelSize_mm'] * Axis_conversion
        #assert not np.isnan(derta_mm).any(), "derta_mm"
        arrow_mm = np.array([derta_mm[0], derta_mm[1], 0]) + arrow_abs_mm
        #assert not np.isnan(arrow_mm).any(), "arrow_mm"
        screen_point_now = Rs2w.dot(arrow_mm) + Ts2w.T
        # import pdb; pdb.set_trace()

        x_data.append([screen_point_now[0]])
        y_data.append([screen_point_now[1]])
        z_data.append([screen_point_now[2]])

    x_data = np.array(x_data)
    y_data = np.array(y_data)
    z_data = np.array(z_data)   
    screen_points = np.column_stack((x_data, y_data, z_data))

    if debug:
        print(" "*100)
        print('*'*100)
        print('screen x:', np.nanmax(x_data), np.nanmin(x_data), np.nanmax(x_data) - np.nanmin(x_data))
        print('screen y:', np.nanmax(y_data), np.nanmin(y_data), np.nanmax(y_data)-np.nanmin(y_data))
        print('screen z:', np.nanmax(z_data), np.nanmin(z_data), np.nanmax(z_data)-np.nanmin(z_data))
        print('*'*100)
        plt.figure()
        plt.scatter(x_data, y_data, alpha=0.5)
        plt.title('screen points')
        plt.xlabel('X Axis')
        plt.ylabel('Y Axis')

        # 调整子图之间的间距
        plt.tight_layout()
        plt.savefig(os.path.join(save_path, str(cam_id) + "_screen_points.png"))
        # 显示图像
        plt.show()
    # import pdb; pdb.set_trace()
    return screen_points

def get_white_mask(white_path, bin_thresh=12):  
    gray = cv2.imread(white_path, cv2.COLOR_BGR2GRAY)
    mask = np.zeros_like(gray)
    _,  thresh_img = cv2.threshold(gray, bin_thresh, 255, cv2.THRESH_BINARY)
    contours, _ = cv2.findContours(thresh_img, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    max_index, max_contour = max(enumerate(contours), key=lambda x: cv2.boundingRect(x[1])[2] * cv2.boundingRect(x[1])[3]) 
    cv2.drawContours(mask, [max_contour], -1, (255, 255, 255), thickness=cv2.FILLED)
    return mask

# def get_world_points_from_mask(binary_image, world_mat_contents, d, save_path, debug=False):

#     # Get the coordinates of non-zero pixels (points inside the mask)
#     mask_points = np.array(np.column_stack(np.where(binary_image > 0)))
#     # Print the number of sampled points
#     num_points = len(mask_points)
#     # print(f"Number of sampled points: {num_points}")
    
#     A = world_mat_contents['camera_matrix']
#     Rw2c = world_mat_contents['rotation_matrix']
#     Rc2w = np.linalg.inv(Rw2c)
#     Tw2c = world_mat_contents['translation_vector']
#     Tc2w = -np.dot(np.linalg.inv(Rw2c), Tw2c)
#     # import pdb; pdb.set_trace()
#     # world_points = affine_img2world(grid_points, A, Rc2w, Tc2w, d)
#     world_points = affine_img2world(mask_points, A, Rc2w, Tc2w, d)

#     if debug:
#         # Plot the original mask and sampled points
#         plt.scatter(world_points[:, 1], world_points[:, 0], color='red', s=1, label="Sampled Points")
#         plt.title(f'Sampled Points from Mask')
#         plt.legend()
#         plt.show()

#     return world_points

def get_world_points_from_mask(grid_points, world_mat_contents, d, save_path, debug=False):

    # Get the coordinates of non-zero pixels (points inside the mask)
    #mask_points = np.array(np.column_stack(np.where(binary_image > 0)))
    # Print the number of sampled points
    # print(f"Number of sampled points: {num_points}")
    
    A = world_mat_contents['camera_matrix']
    Rw2c = world_mat_contents['rotation_matrix']
    Rc2w = np.linalg.inv(Rw2c)
    Tw2c = world_mat_contents['translation_vector']
    Tc2w = -np.dot(np.linalg.inv(Rw2c), Tw2c)
    # import pdb; pdb.set_trace()

    world_points = affine_img2world(grid_points, A, Rc2w, Tc2w, d)

    if debug:
        # Plot the original mask and sampled points
        plt.scatter(world_points[:, 1], world_points[:, 0], color='red', s=1, label="Sampled Points")
        plt.title(f'Sampled Points from Mask')
        plt.legend()
        plt.show()

    return world_points

# 定义旋转矩阵 (例如：绕z轴旋转theta角)
def rotation_matrix_z(theta):
    return np.array([
        [np.cos(theta), -np.sin(theta), 0],
        [np.sin(theta), np.cos(theta), 0],
        [0, 0, 1]
    ])

# 定义平移向量
def translation_vector(tx, ty, tz):
    return np.array([tx, ty, tz])

# 对点云应用旋转和平移
def transform_point_cloud(point_cloud, rotation_matrix, translation_vector):
    # 应用旋转
    rotated_point_cloud = np.dot(point_cloud, rotation_matrix.T)  # 转置矩阵用于点的正确旋转
    # 应用平移
    translated_point_cloud = rotated_point_cloud + translation_vector
    return translated_point_cloud




def post_process(txt_path, debug):
    #txt_path = 'D:\\file\\20240913-data\\20240913105139948\\'
    total_cloud_point = np.empty((0, 3))
    data = []
    for pt_txt_path in  glob.glob(os.path.join(txt_path, "*.txt"))[:4]:
        print(pt_txt_path)
        data.append(np.loadtxt(os.path.join(pt_txt_path), delimiter=','))

    theta0 = np.radians(-90)
    rotation_matrix0 = rotation_matrix_z(theta0)
    translation0 = translation_vector(0, 0, 0)
    data[0] = transform_point_cloud(data[0], rotation_matrix0, translation0)
    theta0 = np.radians(0)
    rotation_matrix0 = rotation_matrix_z(theta0)
    translation0 = translation_vector(0-np.max(data[0][:,0]), (0-np.min(data[0][:,1])), 0)
    data[0] = transform_point_cloud(data[0], rotation_matrix0, translation0)

    theta1 = np.radians(0)
    rotation_matrix1 = rotation_matrix_z(theta1)
    translation1 = translation_vector(0-np.min(data[1][:,0]), (0-np.max(data[1][:,1])), 0)
    data[1] = transform_point_cloud(data[1], rotation_matrix1, translation1)

    theta2 = np.radians(90)
    rotation_matrix2 = rotation_matrix_z(theta2)
    translation2 = translation_vector(0, 0, 0)
    data[2] = transform_point_cloud(data[2], rotation_matrix2, translation2)
    theta2 = np.radians(0)
    rotation_matrix2 = rotation_matrix_z(theta2)
    translation2 = translation_vector(0-np.min(data[2][:,0]), (0-np.min(data[2][:,1])), 0)
    data[2] = transform_point_cloud(data[2], rotation_matrix2, translation2)

    theta3 = np.radians(0)
    rotation_matrix3 = rotation_matrix_z(theta3)
    translation3 = translation_vector(0-np.max(data[3][:,0]), (0-np.max(data[3][:,1])), 0)
    data[3] = transform_point_cloud(data[3], rotation_matrix3, translation3)
    
    x_range = np.max(data[1][:, 0]) - np.min(data[3][:, 0])
    y_range = np.max(data[0][:, 1]) - np.min(data[3][:, 1])
    
    data[0] = transform_point_cloud(data[0], rotation_matrix_z(0), translation_vector(0, (0-(y_range-300)/2), 0))
    data[1] = transform_point_cloud(data[1], rotation_matrix_z(0), translation_vector((0-(x_range-300)/2), 0, 0))
    data[2] = transform_point_cloud(data[2], rotation_matrix_z(0), translation_vector(0, (0-(y_range-300)/2), 0))
    data[3] = transform_point_cloud(data[3], rotation_matrix_z(0), translation_vector((0+(x_range-300)/2), 0, 0))

    for i in range(4):
        # print('i = max x = ',i , np.max(data[i][:, 0]))
        # print('i = min x = ',i , np.min(data[i][:, 0]))
        # print('i = max y = ',i , np.max(data[i][:, 1]))
        # print('i = min y = ',i , np.min(data[i][:, 1]))
        data[i][:, 2] = data[i][:, 2] - np.mean(data[i][:, 2])
        total_cloud_point = np.vstack([total_cloud_point, np.column_stack((data[i][:, 0], data[i][:, 1], data[i][:, 2]))])
    if debug:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        # 提取 x, y, z 坐标
        x_vals = total_cloud_point[:, 0]
        y_vals = total_cloud_point[:, 1]
        z_vals = total_cloud_point[:, 2]

        # 绘制3D点云
        ax.scatter(x_vals, y_vals, z_vals, c=z_vals, cmap='viridis', marker='o')

        # 设置轴标签和标题
        ax.set_xlabel('X (mm)')
        ax.set_ylabel('Y (mm)')
        ax.set_zlabel('Z (um)')
        ax.set_title('3D Point Cloud Visualization')

        plt.show()